"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements CoRL experiments for safe autonomy training
runs.

CoRL `Experiment`s act as containers around training runs. Within
an experiment a user can define custom training behavior and
develop integration for different RL frameworks and backend simulator
solutions.
"""
